-- This file handles the block you're moving and the preview. It
-- defines a class Shape for this purpose.

-- BloxShape class constructor
BloxShape = {} -- static blocks
AnimBlock = {} -- animated blocks
local random = random
local floor = floor
local ExplodingBlockFrames

function BloxShape:new(o)
   local tmp_id = random(BloxConstants.NUM_SHAPES) 
   local tmp_rot = random(4) 
   if not o then
      o = {
	 id = tmp_id, -- shape ID
	 x = floor(BloxGameBoard.WIDTH/2), 
	 rot = tmp_rot, -- rotation
	 y = BloxConstants.BLOX_SHAPE_OFFSETS[tmp_id][tmp_rot], -- y position
	 drop_tick = 1.2, -- s delay between drops
	 next_drop = 0, -- time of next drop
	 next_move = 0, -- time for next horizontal move
	 gfx_id = tmp_id,  --random(BloxConstants.NUM_BLOCK_GFX), -- texture num
	 moved = true, -- needs redraw?
	 x_mov = 0, -- horizontal movement active
      }
   end
   setmetatable(o, self)
   if self.__index ~= self then self.__index = self end
   return o
end

function AnimBlock:new(o)
   if not o then
      o = {
	 dest_x_tile, dest_y_tile, -- destination galore
	 dest_x , dest_y,  -- destination pos (in pixels)
	 pos_x, pos_y,     -- position in pixels
	 delta_y = 0.0,    -- x/y acceleration, in pixels
	 delta_x = 0.0,
	 active = false,   -- is block currently active. 
	 mode = BloxConstants.MODE_FALLING, -- MODE_FALLING, MODE_EXPLODE
	 tile = nil,
	 alpha = 1.0
      }
   end
   setmetatable(o, self)
   if self.__index ~= self then self.__index = self end
   return o
end

function AnimBlock:Update(delta) 
   -- Update animation, return true animating, false false if inactive or done
   if not self.active then
      return false
   end
   --   if(self.delay > 0) then
   --      self.delay = self.delay - BloxFrame.TimeSinceLastUpdate
   --      if self.delay > 0 then
   -- 	 return true
   --      end
   --   end

   --   DEFAULT_CHAT_FRAME:AddMessage(string.format("pos_y = %f, dest_y = %d, delta_y =  %f\n",
   --					       self.pos_y, self.dest_y, self.delta_y))
   if self.mode == BloxConstants.MODE_FALLING then
      self.delta_y = self.delta_y + 0.15 * BloxFrame.TimeSinceLastUpdate*200 -- accel
      self.pos_y = self.pos_y - self.delta_y
       if self.pos_y <= self.dest_y then
	  self.pos_y = self.dest_y
	  self.delta_y = 0
	  self.active = false
       end
       BloxBoardWidgets[self.dest_x_tile][self.dest_y_tile]:SetPoint("BOTTOMLEFT", self.pos_x, self.pos_y)
    elseif self.mode == BloxConstants.MODE_EXPLODE then
       self.delta_y = self.delta_y + 0.15 * BloxFrame.TimeSinceLastUpdate*220 -- accel
       self.pos_y = self.pos_y - self.delta_y
       self.pos_x = self.pos_x - self.delta_x
       self.alpha = self.alpha - BloxFrame.TimeSinceLastUpdate*3
       if(self.alpha < 0) then
	  self.active = false
	  self.tile:SetAlpha(1.0)
	  self.tile:Hide()
       else 
	  self.tile:SetAlpha(self.alpha)
	  self.tile:SetPoint("BOTTOMLEFT", self.pos_x, self.pos_y)
       end
    end

    return self.active
end

-- Returns true if it's time to move the shape down.
function BloxShape:dropTime()
   return BloxFrame.currentTicks >= self.next_drop
end

-- Checks if the x/y position of the current shape
-- is a displayable item.
function BloxShape:isMasked(xpos, ypos)
   return BloxConstants.BLOX_SHAPE_MASKS[self.id][self.rot][xpos][ypos] == 1
end

-- Check if the position of the block is valid.
function BloxShape:isValidPos()
   local orig_x = self.x
   local orig_y = self.y
   local board = BloxGame.board.board
   local mx = BloxGameBoard.WIDTH
   local my = BloxGameBoard.HEIGHT
   local isMasked = self.isMasked
   for i = 1, 4 do
      for j = 1, 4 do
	 if isMasked(self, i, j) then 
	    local xi = orig_x + i - 1
	    local yj = orig_y + j - 1
	    if xi < 1 or xi > mx  or
	       yj < 1 or yj > my or
	       board[yj][xi] > 0 -- not IsTileEmpty
	    then
	       return false
	    end
	 end
      end
   end
   return true
end

-- Move down once
function BloxShape:moveDown()
   self.y = self.y + 1
   if not self:isValidPos() then
      self.y = self.y - 1
      return false
   end
   self.next_drop = BloxFrame.currentTicks + self.drop_tick
   self.moved = true
   return true
end

-- Move the block left, if possible.
function BloxShape:moveLeft()
   local orig_x = self.x
   self.x = self.x - 1
   if not self:isValidPos() then
      self.x = orig_x
      return false
   end
   self.moved = true
   return true
end

-- Move the block right, if possible.
function BloxShape:moveRight()
   local orig_x = self.x
   self.x = self.x + 1
   if not self:isValidPos() then
      self.x = orig_x
      return false
   end
   self.moved = true
   return true
end

-- This rotates the shape, if possible. 
function BloxShape:rotate()
   --   self:debug()
   self.rot = (self.rot + 1)
   if self.rot > 4 then self.rot = 1 end
   if not self:isValidPos() then
      if self.y <= 1 then
	 -- Attempt to move down once to see if it works.
	 self.y = self.y + 1
	 if self:isValidPos() then
	    self.moved = true
	    return true
	 else
	    self.y = self.y - 1
	 end
      end
      self.rot = self.rot - 1
      if self.rot < 1 then self.rot = 4 end
      return false
   end
   self.moved = true
   return true
end

-- Drop the shape to the bottom
function BloxShape:drop(bottom)
--   dbg_()
   local board = BloxGame.board
   local isMasked = self.isMasked
   if not bottom then
      while self:moveDown() do
	 -- Empty
      end
   end
   for i = 1, 4 do
      for j = 1, 4 do
	 local xi = self.x + i -1
	 local yj = self.y + j - 1
	 if isMasked(self, i, j) and (xi > 0) and (yj > 0) then
	    BloxShapeWidgets[i][j]:Hide()
	    board:SetTile(xi, yj, self.gfx_id)
	 end
      end
   end   
   
   -- Initialize to the value of the next shape.
   local next = board.next
   local new_texture = self.gfx_id ~= next.gfx_id
   self.x      = next.x
   self.y      = next.y
   self.rot    = next.rot
   self.id     = next.id
   self.gfx_id = next.gfx_id

   -- Reset the next shape 
   --   next:randomize()
   next.moved = true

   -- Reset other data
   self.next_drop = BloxFrame.currentTicks + self.drop_tick
   self.moved = true

   -- Update texture for the new gfx_id if needed
   if new_texture then self:fixTextures() end

   -- Check for lines to remove
   board:CheckLines()
   board:CheckGameOver()
end

function BloxShape:fixTextures(next) 
   local tiles
   if next then
      tiles = BloxNextShapeWidgets
   else
      tiles = BloxShapeWidgets
   end
   for x=1, 4 do
      for y=1, 4 do
	 tiles[x][y].backdrop.bgFile = BloxConstants.TEXTURES[self.gfx_id]
	 tiles[x][y]:SetBackdrop(tiles[x][y].backdrop);
--	 tiles[x][y]:SetNormalTexture( BloxConstants.TEXTURES[self.gfx_id] )
      end
   end
end

-- Print debug info
function BloxShape:debug()
   local info = self:shapeInfo()
   DEFAULT_CHAT_FRAME:AddMessage(string.format("DEBUG: id=%.2f, x=%.2f, y=%.2f, rot=%.2f, drop_tick=%.2f, next_drop=%.2f, next_move=%.2f, gfx_id=%.2f, si=%d, %d, %d, %d\n", 
					       self.id, self.x, self.y, self.rot, self.drop_tick,
					       self.next_drop, self.next_move,
					       self.gfx_id,
					       info[1], info[2], info[3], info[4]))
   self:dumpShape(self)
end

-- Debug function. Draw the current shape in the chat window.
function BloxShape:dumpShape(shape) 
   local dbg = ""
   for y =1, 4 do
      for x = 1, 4 do
	 if shape:isMasked(x, y) then
	    dbg = dbg .. "* "
	 else
	    dbg = dbg .. "  "
	 end
      end
      dbg = dbg .. "\n"
   end
   DEFAULT_CHAT_FRAME:AddMessage(dbg)
end

-- Toggle to the next shape type.
function BloxShape:nextType()
   self.id = self.id + 1
   if self.id > BloxConstants.MAX_SHAPE_ID then
      self.id = 1
   end
   self.moved = true
end

-- Randomize color and shape, reset variables
function BloxShape:randomize()
   self.id = random(BloxConstants.NUM_SHAPES)
   self.rot = random(4)
   self.x = floor((BloxGameBoard.WIDTH - self:shapeInfo()[2] +
			self:shapeInfo()[4]) / 2)+1
   self.y = BloxConstants.BLOX_SHAPE_OFFSETS[self.id][self.rot]
   self.next_drop = BloxFrame.currentTicks + self.drop_tick
   self.gfx_id = self.id --random(BloxConstants.NUM_BLOCK_GFX)
   self.moved = true
   self:fixTextures(true)
end

-- Set the delay between each vertical drop
function BloxShape:setDropTick(_drop_tick)
   if self.drop_tick ~= _drop_tick then
      self.drop_tick = _drop_tick
      self.next_drop = BloxFrame.currentTicks + self.drop_tick
   end
end

-- Return info about current shape
function BloxShape:shapeInfo()
   return BloxConstants.BLOX_SHAPE_MASKS[self.id][self.rot][5]
end

-- Move the block left or right.
function BloxShape:moveHorizontal()
   if self.x_mov == 0 or BloxFrame.currentTicks < self.next_move then
      return
   end
   self.next_move = BloxFrame.currentTicks + 0.04
   if self.x_mov < 0 then
      self:moveLeft()
   else
      self:moveRight()
   end
end

-- Start horizontal movement. dir = -1  => left, 1 = right
function BloxShape:setMove(dir)
   self.x_mov = dir
   self.next_move = 0
   self:moveHorizontal() -- force a move
   
   -- Makes the next horizontal movement delayed 200 ms
   -- rather than the normal 40.
   self.next_move = self.next_move + 0.16
end

function AnimBlock.CreateExplodingFrames()
   -- Create the frames for the exploding blocks.
   ExplodingBlockFrames = { }
   for mx = 1, BloxGameBoard.WIDTH do
      ExplodingBlockFrames[mx] = {}
      for my = 1, 4 do
	 f = CreateFrame("Frame", nil, BloxGameBoard)
	 f:SetWidth(19)	 
	 f:SetHeight(19)
	 ExplodingBlockFrames[mx][my] = f
	 f.backdrop = { tile = false }
	 f:SetBackdrop(f.backdrop)
      end
   end
end

function AnimBlock.Create(board, anim, sx, sy, x, y, mode, delay) 
   anim.pos_x, anim.pos_y = board.GetPixelPosition(sx, sy)   
   anim.dest_x, anim.dest_y = board.GetPixelPosition(x, y)
   if mode == BloxConstants.MODE_FALLING then
      anim.delta_y = 0.0
      anim.delta_x = 0.0
      BloxBoardWidgets[x][y]:SetPoint("BOTTOMLEFT", anim.pos_x, anim.pos_y)
   else
      anim.delta_y = 1 - random(4.0)
      anim.delta_x =  BloxGameBoard.WIDTH/3.0 - sx/1.5  + 1.5 - random(3)
      anim.alpha = 1.0
      if not anim.tile then
	 if ExplodingBlockFrames == nil then
	    AnimBlock.CreateExplodingFrames()
	 end
	 anim.tile = ExplodingBlockFrames[x][y]
      end
      anim.tile.backdrop.bgFile = BloxConstants.TEXTURES[BloxGame.board.board[sy][sx]];
      anim.tile:SetBackdrop(anim.tile.backdrop);
      anim.tile:Show()
      BloxGame.board:AdjustBlock(f, sx, sy, true, false)
   end
   
   anim.dest_x_tile = x
   anim.dest_y_tile = y
   anim.mode = mode
   anim.active = true
   anim.delay = delay
end
